import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class sketch extends PApplet {

// Data containers
Table data;
int rows,cols;
int hiddenCol;
int[][] dataMatrix;
int[] maxMatrix;
int[] minMatrix;
Cell[][] cellMatrix;

// Layout parameters
final int cellWidthMax = 100;
final int cellWidthMin = 10;
final int cellHeight = 4;
final int cellXDist = 1;
final int cellYDist = 2;
final int sideDist = 60;
final int middleDist = 30;
final int fontYOffset = -35;
final float labelRelX = 1.7f;
final float labelRelY = 2.0f;
final int switchSize = 10;
int firstTwo;

// Color parameters
final int bgColor = 255;
final int bgFontColor = 200;
final int labelFontColor = 180;

// Misc.
final int absMax = 2147483647;
final int uDefI = 0;
final boolean uDefB = false;

// Interface
Checkbox checkboxDeweyWarp;
boolean deweyWarp = true;
PFont labelFont,bgFont;
String bgValUpper = "";
String bgValLower = "";


public void setup() {
  
  smooth();

  // Set font
  labelFont = createFont("Futura",10);
  bgFont = createFont("Futura",280);
  
  // Load data
  data = loadTable("assignment3_data.csv", "header");
  rows = data.getRowCount();
  cols = data.getColumnCount();
  
  // Hidden column is the last column
  hiddenCol = cols-1;
  
  // Resize arrays
  dataMatrix = new int[rows][cols];
  maxMatrix = new int[cols];
  minMatrix = new int[cols];
  cellMatrix = new Cell[rows][cols];

  // Populate data matrix
  for (int i=0; i<rows; i++) {
    for (int j=0; j<cols; j++) {
      dataMatrix[i][j] = data.getInt(i,j);
    }
  }

  // Prepare minimum and maximum matrices
  for (int i=0; i<rows; i++) {
    for (int j=0; j<cols; j++) {
      maxMatrix[j] = 0;
      minMatrix[j] = absMax;  
    }
  }

  // Populate minimum and maximum matrices
  for (int i=0; i<rows; i++) {
    for (int j=0; j<cols; j++) {
      if (dataMatrix[i][j] < minMatrix[j]) minMatrix[j] = dataMatrix[i][j];
      if (dataMatrix[i][j] > maxMatrix[j]) maxMatrix[j] = dataMatrix[i][j];
    }
  }

  // Populate cell matrix
  for (int i=0; i<rows; i++) {
    for (int j=0; j<cols; j++) {
      cellMatrix[i][j] = new Cell(uDefI,uDefI,uDefI,cellHeight,dataMatrix[i][j],minMatrix[j],maxMatrix[j],uDefB,uDefB,cellWidthMax);
    }
  }

  // Cut the hidden column
  cols-=1;
  
  // Dynamic size, first two columns are always minimum width and followed by middleDist
  firstTwo = sideDist+2*(cellWidthMin+cellXDist)+middleDist;
  size(firstTwo+(cols-2)*(cellWidthMax+cellXDist)+sideDist,rows*(cellHeight+cellYDist)+2*sideDist);
  
  // Checkbox for Dewey warping
  checkboxDeweyWarp = new Checkbox(firstTwo,height-(sideDist/2+switchSize/2),switchSize,switchSize,labelFontColor,"DEWEY WARP ON","DEWEY WARP OFF");
  
};


public void draw() {
  
  background(255);
  
  // Draw the background values, position includes hardcoded offset
  textFont(bgFont);
  textAlign(CENTER,CENTER);
  fill(bgFontColor);
  text(bgValUpper,width/2,(height/3)*1+fontYOffset);
  text(bgValLower,width/2,(height/3)*2+fontYOffset);
  
  // Reset background values in each draw cycle
  bgValUpper = "";
  bgValLower = "";
  
  for (int i=0; i<rows; i++) {
    for (int j=0; j<cols; j++) {

      boolean inFirstTwo = false;
      if (j==0 || j==1) {inFirstTwo = true;}
        
      // If we're not in the first two columns and Dewey warping is on, use relative width
      if (!inFirstTwo && deweyWarp) {
        cellMatrix[i][j].w = PApplet.parseInt(map(PApplet.parseFloat(dataMatrix[i][hiddenCol]),PApplet.parseFloat(minMatrix[hiddenCol]),PApplet.parseFloat(maxMatrix[hiddenCol]),cellWidthMin,cellWidthMax));
        cellMatrix[i][j].x = firstTwo+(j-2)*(cellWidthMax+cellXDist);
      }
      // If we're not in the first two columns and Dewey warping is off, use maximum width
      else if (!inFirstTwo && !deweyWarp) {
        cellMatrix[i][j].w = cellWidthMax;
        cellMatrix[i][j].x = firstTwo+(j-2)*(cellWidthMax+cellXDist);
      }
      // If we're in the first two columns and Dewey warping is on or off, use minimum width
      else {
        cellMatrix[i][j].w = cellWidthMin;
        cellMatrix[i][j].x = sideDist+j*(cellWidthMin+cellXDist);
      }
        
      cellMatrix[i][j].y = sideDist+i*(cellHeight+cellYDist);
      cellMatrix[i][j].v = dataMatrix[i][j];
        
      if (cellMatrix[i][j].mouseOver()) {
        bgValUpper = str(dataMatrix[i][j]);       
        // Show the lower values and the maximum outline only if we're not in the first two columns
        if (!inFirstTwo) {
          bgValLower = str(dataMatrix[i][hiddenCol]);
          cellMatrix[i][j].showMaxOutline = true;
        }
        else {
          cellMatrix[i][j].showMaxOutline = false;    
        }
      }

      cellMatrix[i][j].draw();
  
    }
  }
  
  textFont(labelFont);
  fill(labelFontColor);
  
  int labelOffset;
  if (deweyWarp) {
    textAlign(LEFT,CENTER);
    labelOffset = firstTwo;
  }
  else {
    textAlign(CENTER,CENTER);
    labelOffset = firstTwo+cellWidthMax/2;
  }
  
  // Horizontal labels
  text("MARX", labelOffset,sideDist/labelRelX);
  text("KEYNES", labelOffset+1*(cellWidthMax+cellXDist),sideDist/labelRelX);
  text("CAPITAL", labelOffset+2*(cellWidthMax+cellXDist),sideDist/labelRelX);
  text("CRISIS", labelOffset+3*(cellWidthMax+cellXDist),sideDist/labelRelX);
  
  // Vertical labels
  pushMatrix();
    translate(sideDist/labelRelY,height/2);
    rotate(radians(270));
    textAlign(CENTER,CENTER);
    text("FOUR RECESSION RELATED KEYWORDS IN TITLES 2006-2014 AND TOTAL TITLES FOR DEWEY DECIMALS 330-339: ECONOMICS",0,0);
  popMatrix();
  
  checkboxDeweyWarp.draw();
  
};


public void mouseClicked() {
  deweyWarp = checkboxDeweyWarp.mouseClicked();
};


// Java passes arrays as reference!
public void bubbleSort(int[][] a, int c) {

  boolean s = true;

  while (s == true) {
    s = false;
    for (int n=0; n<rows-1; n++) {
      // Compare the values in just one column, but swap values in all columns, including the hidden column
      if (a[n][c] > a[n+1][c]) {
        for (int r=0; r<cols+1; r++) {
          int t = a[n][r];
          a[n][r] = a[n+1][r];
          a[n+1][r] = t;
        }
        s = true;
      }
    }
  }

};


public void keyPressed() {

  if (key == '1') {bubbleSort(dataMatrix,0);}
  if (key == '2') {bubbleSort(dataMatrix,1);}
  if (key == '3') {bubbleSort(dataMatrix,2);}
  if (key == '4') {bubbleSort(dataMatrix,3);}
  if (key == '5') {bubbleSort(dataMatrix,4);}
  if (key == '6') {bubbleSort(dataMatrix,5);}
  if (key == '7') {bubbleSort(dataMatrix,6);}

};
class Cell {

  int x,y,h;
  int w,oldW;
  int v,oldV,vMin,vMax;
  int currentC,currentW;
  boolean showOutline = false;
  boolean showMaxOutline = false;
  int maxOutline;
  int minOutline;

  boolean aniVRunning = false;
  boolean aniWRunning = false;
  float aniVDiff,aniV,aniWDiff,aniW;
  int aniVDir,newAniV,aniWDir,newAniW;
  final float aniSpeed = 50;
  
  final int outlineColor = 180;
  final float fillGradientAngle = 97.7f;
  final int gradientMin = 2;
  final int gradientMax = 255;

  Cell(int pX, int pY, int pW, int pH, int pV, int pVMin, int pVMax, boolean pShowOutline, boolean pShowMaxOutline, int pMaxOutline) {
    x = pX;
    y = pY;
    w = pW;
    h = pH;
    v = pV;
    vMin = pVMin;
    vMax = pVMax;
    showOutline = pShowOutline;
    showMaxOutline = pShowMaxOutline;
    maxOutline = pMaxOutline;
    minOutline = w;
  }

  public void draw() {

    animate();

    pushStyle();
      
      noStroke();
      colorMode(HSB);
      fill(map(fillGradientAngle,0,360,0,255),currentC,255);
      rect(x,y,currentW,h);

      if (showOutline) {               
        stroke(outlineColor);
        noFill();
        if (showMaxOutline) {rect(x,y,maxOutline,h);}
        else {rect(x,y,minOutline,h);} 
      }

    popStyle();

  };

  public void animate() {

    // Has the v value changed?
    if (v != oldV) {
      // To avoid jumps when animation is already running
      if (aniVRunning) {oldV = PApplet.parseInt(aniV);}
      aniV = oldV;
      newAniV = v;
      aniVDiff = v-oldV;
      if (aniVDiff > 0) aniVDir = 1;
      if (aniVDiff < 0) aniVDir = -1;
      aniVRunning = true;
    }

    // Has the w value changed?
    if (w != oldW) {
      // To avoid jumps when animation is already running
      if (aniWRunning) {oldW = PApplet.parseInt(aniW);}
      aniW = oldW;
      newAniW = w;
      aniWDiff = w-oldW;
      if (aniWDiff > 0) aniWDir = 1;
      if (aniWDiff < 0) aniWDir = -1;
      aniWRunning = true;
    }

    if (aniVRunning) {
      if (aniV >= newAniV && aniVDir == 1 || (aniV <= newAniV && aniVDir == -1))  {aniVRunning = false;}
      aniV += aniVDiff/aniSpeed;
      currentC = PApplet.parseInt(map(aniV,PApplet.parseFloat(vMin),PApplet.parseFloat(vMax),gradientMin,gradientMax));
    }
    if (!aniVRunning) {
      currentC = PApplet.parseInt(map(PApplet.parseFloat(v),PApplet.parseFloat(vMin),PApplet.parseFloat(vMax),gradientMin,gradientMax));
    }

    if (aniWRunning) {
      if (aniW >= newAniW && aniWDir == 1 || (aniW <= newAniW && aniWDir == -1))  {aniWRunning = false;}
      aniW += aniWDiff/aniSpeed;
      currentW = PApplet.parseInt(aniW);
    }
    if (!aniWRunning) {
      currentW = w;
    }

    oldV = v;
    oldW = w;

  };

  public boolean mouseOver() {
    if (mouseX > x && mouseX < x+w && mouseY > y && mouseY < y+h) {
      showOutline = true;
      return true;
    }
    else {
      showOutline = false;
      return false;
    }
  }

};


class Checkbox {

  int x,y;
  int w = 10;
  int h = 10;
  boolean status = true;
  int c = color(128);
  String labelOn = "ON";
  String labelOff = "OFF";
  final int textOffset = -1;
  final int labelDistance = 10;

  Checkbox(int pX,int pY,int pW, int pH, int pC, String pLabelOn, String pLabelOff) {
    x = pX;
    y = pY;
    w = pW;
    h = pH;
    c = pC;
    labelOn = pLabelOn;
    labelOff = pLabelOff;
  };

  public void draw() {
    pushStyle(); 
      stroke(c);
      noFill();
      textAlign(LEFT,TOP);
      rect(x,y,w,h);
      if (status) {
        line(x,y,x+w,y+h);
        line(x+w,y,x,y+h);
        text(labelOn,x+w+labelDistance,y+textOffset);
      }
      else {
        text(labelOff,x+w+labelDistance,y+textOffset);
      }
    popStyle();
  };

  public boolean mouseClicked() {
    if (mouseX > x && mouseX < x+w && mouseY > y && mouseY < y+h) {
      status = !status;
    }
    return status;
  };

};
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "sketch" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
